from antakia_core.compute.model_subtitution.model_class import MLModel, LinearMLModel, AvgRegressionBaseline
from sklearn import linear_model
from sklearn import tree
import pygam
from interpret.glassbox import ExplainableBoostingRegressor


####################################
# Baseline Models                  #
####################################

class AvgBaselineModel(MLModel):
    def __init__(self):
        super().__init__(AvgRegressionBaseline(), 'average baseline')


####################################
# Linear Models                    #
####################################
class LinearRegression(LinearMLModel):
    def __init__(self):
        super().__init__(linear_model.LinearRegression(), 'linear regression')


class LassoRegression(LinearMLModel):
    def __init__(self):
        super().__init__(linear_model.LassoCV(), 'Lasso regression')


class RidgeRegression(LinearMLModel):
    def __init__(self):
        super().__init__(linear_model.RidgeCV(), 'Ridge regression')


####################################
# DecisionTree Models              #
####################################

class DecisionTreeRegressor(MLModel):
    def __init__(self):
        super().__init__(tree.DecisionTreeRegressor(max_depth=7), 'Decision Tree')


####################################
# GAM Models                       #
####################################

class GaM(MLModel):
    def __init__(self):
        super().__init__(pygam.pygam.GAM(distribution='normal', link='identity'), 'linear GAM')


####################################
# GAM Models                       #
####################################
class EBM(MLModel):
    def __init__(self):
        super().__init__(ExplainableBoostingRegressor(), 'Explainable Boosting Tree')



from __future__ import annotations

import numpy as np
import pandas as pd

import os
from antakia_core.utils.utils import boolean_mask, format_number
from antakia_core.utils.variable import Variable


class Rule:
    """
    class to represent logical rule over a variable
    """

    def __init__(self, variable: Variable, min: float | None = None, includes_min: bool | None = None,
                 max: float | None = None, includes_max: bool | None = None, cat_values: list | set | None = None):
        self.variable = variable
        if cat_values is not None:
            self.cat_values = set(cat_values)
            self.min = None
            self.max = None
            self.includes_min = None
            self.includes_max = None
            self.categorical_rule = True
        else:
            self.categorical_rule = False
            self.cat_values = None
            if min is None:
                self.min = -np.inf
                self.includes_min = False
            else:
                self.min = min
                self.includes_min = includes_min if includes_min else False
            if max is None:
                self.max = np.inf
                self.includes_max = False
            else:
                self.max = max
                self.includes_max = includes_max if includes_max else False

    def __eq__(self, other):
        if self.rule_type != other.rule_type or self.variable != other.variable:
            return False
        if self.rule_type in (-1, 5):
            return True
        if self.rule_type == 0:
            return len(self.cat_values.symmetric_difference(other.cat_values)) == 0
        if self.rule_type == 4:
            return self.min == other.min
        if self.rule_type == 1:
            return self.max == other.max and self.includes_max == other.includes_max
        if self.rule_type == 2:
            return self.min == other.min and self.includes_min == other.includes_min
        return (
            self.max == other.max and self.includes_max == other.includes_max
        ) and (
            self.min == other.min and self.includes_min == other.includes_min
        )

    def __repr__(self):
        if self.is_categorical_rule:
            txt = f"{self.variable.display_name} \u2208  \u27E6"
            txt += ', '.join(self.cat_values)
            txt += "\u27E7"
            return txt
        if self.rule_type == -1:
            return f'{self.variable.display_name} - True'
        if self.rule_type == 5:
            return f'{self.variable.display_name} - False'
        if self.rule_type == 1:
            # Rule type 1
            op = '\u2264' if self.includes_max else '<'
            txt = f"{self.variable.display_name} {op} {format_number(self.max)}"
            return txt
        if self.rule_type == 2:
            # Rule type 2
            op = '\u2265' if self.includes_min else '>'
            txt = f"{self.variable.display_name} {op} {format_number(self.min)}"
            return txt
        if self.rule_type == 3:
            # Rule type 3 : the rule is of the form : variable included in [min, max] interval, or min < variable < max
            if os.environ.get("USE_INTERVALS_FOR_RULES"):
                open_bracket = '\u27E6' if self.includes_min else '['
                close_bracket = '\u27E7' if self.includes_min else ']'
                txt = f"{self.variable.display_name} \u2208 {open_bracket} {format_number(self.min)},"
                txt += f" {format_number(self.max)} {close_bracket}"  # element of
                return txt
            min_op = '\u2264' if self.includes_min else '<'
            max_op = '\u2264' if self.includes_max else '<'

            txt = f'{format_number(self.min)} {min_op} {self.variable.display_name} '
            txt += f'{max_op} {format_number(self.max)}'
            return txt
        # Rule type 4 : the rule is of the form : variable not included in [min, max] interval or variable < min and variable > max
        return f'{self.variable.display_name} = {self.min}'

    @property
    def rule_type(self):
        """
        return type of rule :
            -1: no rule
            0: categorical rule
            1: x <(=) max
            2: x >(=) min
            3: min <(=) x <(=) max
            4: x = value
            5: Falsy rule
        """
        if self.is_categorical_rule:
            if self.cat_values:
                return 0
            else:
                return 5
        if self.min == -np.inf and self.max == np.inf:
            return -1
        if self.min == -np.inf:
            return 1
        if self.max == np.inf:
            return 2
        if self.max > self.min:
            return 3
        if self.max == self.min and self.includes_max and self.includes_min:
            return 4
        return 5

    @property
    def operator_max(self):
        if self.rule_type == 4:
            return '__eq__'
        if self.includes_max:
            return '__le__'
        return '__lt__'

    @property
    def operator_min(self):
        if self.rule_type == 4:
            return '__eq__'
        if self.includes_min:
            return '__ge__'
        return '__gt__'

    @property
    def is_categorical_rule(self):
        return self.categorical_rule

    def get_matching_mask(self, X: pd.DataFrame) -> pd.Series:
        col = X.loc[:, self.variable.column_name]
        if self.rule_type == -1:
            return boolean_mask(X, True)
        if self.rule_type == 0:
            return col.isin(self.cat_values)
        if self.rule_type == 4:
            return col == self.min
        if self.rule_type == 5:
            return boolean_mask(X, False)
        return getattr(col, self.operator_max)(self.max) & getattr(col, self.operator_min)(self.min)

    def __call__(self, value: float | pd.DataFrame) -> bool | pd.Series:
        if isinstance(value, pd.DataFrame):
            return self.get_matching_mask(value)
        if self.rule_type == -1:
            return True
        if self.rule_type == 0:
            return value in self.cat_values
        if self.rule_type == 4:
            return value == self.min
        if self.rule_type == 5:
            return False
        return getattr(value, self.operator_max)(self.max) & getattr(value, self.operator_min)(self.min)

    def combine(self, rule: Rule):
        """
        builds a new combining current rule and argument rules (logical and)
        Parameters
        ----------
        rule

        Returns
        -------

        """

        if self.variable != rule.variable:
            raise ValueError('cannot combine two rules on different variables')
        # categorical rules
        if self.categorical_rule and rule.categorical_rule:
            common_values = set(self.cat_values).intersection(rule.cat_values)
            return Rule(self.variable, cat_values=common_values)
        if rule.categorical_rule:
            return rule.combine(self)
        if self.categorical_rule:
            return Rule(self.variable, cat_values=[v for v in self.cat_values if rule(v)])
        # edge cases
        if self.rule_type == 5 or rule.rule_type == 5:
            return self.copy()
        if self.rule_type == -1:
            return rule.copy()
        if rule.rule_type == -1:
            return self.copy()
        # nominal
        min_val = max(self.min, rule.min)
        if self.min > rule.min:
            include_min = self.includes_min
        elif rule.min > self.min:
            include_min = rule.includes_min
        else:
            include_min = min(self.includes_min, rule.includes_min)

        max_val = min(self.max, rule.max)
        if self.max < rule.max:
            include_max = self.includes_max
        elif rule.max < self.max:
            include_max = rule.includes_max
        else:
            include_max = min(self.includes_max, rule.includes_max)
        return Rule(self.variable, min_val, include_min, max_val, include_max)

    def __and__(self, other):
        return self.combine(other)

    def to_dict(self):
        return {
            'Variable': self.variable.display_name,
            'Unit': self.variable.unit,
            'Desc': self.variable.descr,
            'Critical': self.variable.critical,
            'Rule': self.__repr__()
        }

    def copy(self):
        return Rule(self.variable, self.min, self.includes_min, self.max, self.includes_max, self.cat_values)


class TruthyRule(Rule):
    def __init__(self, var: Variable):
        super().__init__(var)


class FalsyRule(Rule):
    def __init__(self, var: Variable):
        super().__init__(var, min=5, max=4)



from antakia_core.compute.model_subtitution.model_class import MLModel, LinearMLModel, AvgClassificationBaseline
from sklearn import linear_model, tree



####################################
# Baseline Models                  #
####################################

class AvgClassificationBaselineModel(MLModel):
    def __init__(self):
        super().__init__(AvgClassificationBaseline(), 'average baseline')


####################################
# Linear Models                    #
####################################

class LogisticRegression(LinearMLModel):
    def __init__(self):
        super().__init__(linear_model.LogisticRegression(), 'linear regression')


class DecisionTreeClassifier(MLModel):
    def __init__(self):
        super().__init__(tree.DecisionTreeClassifier(), 'Decision Tree')



from collections import namedtuple
import pandas as pd

from antakia_core.compute.dim_reduction.dim_reduc_method import DimReducMethod
from antakia_core.compute.dim_reduction.dim_reduction import compute_projection, dim_reduc_factory

Proj = namedtuple('Proj', ['reduction_method', 'dimension'])


class ProjectedValues:
    def __init__(self, X: pd.DataFrame, y: pd.Series):
        self.X = X
        self.y = y
        self._projected_values = {}
        self._parameters = {}

    def set_parameters(self, projection: Proj, parameters: dict):
        """
        set new parameters for a (projection method, dimension)
        remove previously computed projected value
        Parameters
        ----------
        projection_method : projection method value
        dimension: dimension
        parameters: new parameters

        Returns
        -------

        """
        assert projection.reduction_method in DimReducMethod.dimreduc_methods_as_list()
        assert projection.dimension in [2, 3]

        if self._parameters.get(projection) is None:
            self.build_default_parameters(projection)

        self._parameters[projection]['previous'] = \
            self._parameters[projection]['current'].copy()
        self._parameters[projection]['current'].update(parameters)
        del self._projected_values[projection]

    def get_parameters(self, projection: Proj):
        """
        get the value of the parameters for a (projection method, dimension)
        build it to default if needed
        Parameters
        ----------
        projection_method
        dimension

        Returns
        -------

        """
        if self._parameters.get(projection) is None:
            self.build_default_parameters(projection)
        return self._parameters.get(projection)

    def build_default_parameters(self, projection: Proj):
        """
        build default parameters from DimReductionMethod.parameters()
        Parameters
        ----------
        projection_method
        dimension

        Returns
        -------

        """
        current = {}
        dim_reduc_parameters = dim_reduc_factory[projection.reduction_method].parameters()
        for param, info in dim_reduc_parameters.items():
            current[param] = info['default']
        self._parameters[projection] = {
            'current': current,
            'previous': current.copy()
        }

    def get_projection(self, projection: Proj, progress_callback: callable = None):
        """
        get a projection value
        computes it if necessary
        Parameters
        ----------
        projection_method
        dimension
        progress_callback

        Returns
        -------

        """
        if not self.is_present(projection):
            self.compute(projection, progress_callback)
        return self._projected_values[projection]

    def is_present(self, projection: Proj) -> bool:
        """
        tests if the projection is already computed
        Parameters
        ----------
        projection_method
        dimension

        Returns
        -------

        """
        return self._projected_values.get(projection) is not None

    def compute(self, projection: Proj, progress_callback: callable):
        """
        computes a projection and store it
        Parameters
        ----------
        projection_method
        dimension
        progress_callback

        Returns
        -------

        """
        projected_values = compute_projection(
            self.X,
            self.y,
            projection.reduction_method,
            projection.dimension,
            progress_callback,
            **self.get_parameters(projection)['current']
        )
        self._projected_values[projection] = projected_values



import lime
import numpy as np
import pandas as pd
import shap

from antakia_core.explanation.explanation_method import ExplanationMethod
from antakia_core.utils.utils import ProblemCategory



# ===========================================================
#              Explanations implementations
# ===========================================================


class SHAPExplanation(ExplanationMethod):
    """
    SHAP computation class.
    """

    def __init__(self, X: pd.DataFrame, model, task_type, progress_updated: callable = None):
        super().__init__(ExplanationMethod.SHAP, X, model, task_type, progress_updated)

    @property
    def link(self):
        if self.task_type == ProblemCategory.regression:
            return "identity"
        return "logit"

    def compute(self) -> pd.DataFrame:
        self.publish_progress(0)
        try:
            explainer = shap.TreeExplainer(self.model)
        except:
            explainer = shap.KernelExplainer(self.model.predict, self.X.sample(min(200, len(self.X))), link=self.link)
        chunck_size = 200
        shap_val_list = []
        for i in range(0, len(self.X), chunck_size):
            explanations = explainer.shap_values(self.X.iloc[i:i + chunck_size])
            if isinstance(explanations, list):
                # classification, use only class 1 probabilities
                explanations = explanations[-1]
            shap_val_list.append(
                pd.DataFrame(explanations, columns=self.X.columns, index=self.X.index[i:i + chunck_size]))
            self.publish_progress(int(100 * (i * chunck_size) / len(self.X)))
        shap_values = pd.concat(shap_val_list)
        self.publish_progress(100)
        return shap_values


class LIMExplanation(ExplanationMethod):
    """
    LIME computation class.
    """

    def __init__(self, X: pd.DataFrame, model, task_type, progress_updated: callable = None):
        super().__init__(ExplanationMethod.LIME, X, model, task_type, progress_updated)

    @property
    def mode(self):
        if self.task_type == ProblemCategory.regression:
            return 'regression'
        else:
            return 'classification'

    def compute(self) -> pd.DataFrame:
        self.publish_progress(0)

        explainer = lime.lime_tabular.LimeTabularExplainer(
            self.X.sample(min(len(self.X), 500)).values,
            feature_names=self.X.columns,
            verbose=False,
            mode=self.mode,
            discretize_continuous=False
        )

        values_lime = pd.DataFrame(
            np.zeros(self.X.shape),
            index=self.X.index,
            columns=self.X.columns
        )
        progress = 0
        if self.mode == 'regression':
            predict_fct = self.model.predict
            i = 0
        else:
            i = 1
            if hasattr(self.model, 'predict_proba'):
                predict_fct = self.model.predict_proba
            else:
                predict_fct = self.model.predict
        for index, row in self.X.iterrows():
            exp = explainer.explain_instance(row.values, predict_fct)

            values_lime.loc[index] = pd.Series(exp.local_exp[i], index=explainer.feature_names).str[1][
                values_lime.columns].values
            progress += 100 / len(self.X)
            self.publish_progress(int(progress))
        self.publish_progress(100)
        return values_lime


def compute_explanations(X: pd.DataFrame, model, explanation_method: int, task_type,
                         callback: callable) -> pd.DataFrame:
    """ Generic method to compute explanations, SHAP or LIME
    """
    if explanation_method == ExplanationMethod.SHAP:
        return SHAPExplanation(X, model, task_type, callback).compute()
    elif explanation_method == ExplanationMethod.LIME:
        return LIMExplanation(X, model, task_type, callback).compute()
    else:
        raise ValueError(f"This explanation method {explanation_method} is not valid!")



import pandas as pd

from antakia_core.utils.long_task import LongTask


class ExplanationMethod(LongTask):
    """
    Abstract class (see Long Task) to compute explaination values for the Explanation Space (ES)

    Attributes
    model : the model to explain
    explanation_method : SHAP or LIME
    """

    # Class attributes
    NONE = 0  # no explanation, ie: original values
    SHAP = 1
    LIME = 2

    def __init__(
            self,
            explanation_method: int,
            X: pd.DataFrame,
            model,
            task_type,
            progress_updated: callable = None,
    ):
        if not ExplanationMethod.is_valid_explanation_method(explanation_method):
            raise ValueError(explanation_method, " is a bad explanation method")
        self.explanation_method = explanation_method
        super().__init__(X, progress_updated)
        self.task_type = task_type
        self.model = model

    @staticmethod
    def is_valid_explanation_method(method: int) -> bool:
        """
        Returns True if this is a valid explanation method.
        """
        return (
                method == ExplanationMethod.SHAP
                or method == ExplanationMethod.LIME
                or method == ExplanationMethod.NONE
        )

    @staticmethod
    def explanation_methods_as_list() -> list:
        return [ExplanationMethod.SHAP, ExplanationMethod.LIME]

    @staticmethod
    def explain_method_as_str(method: int) -> str:
        if method == ExplanationMethod.SHAP:
            return "SHAP"
        elif method == ExplanationMethod.LIME:
            return "LIME"
        else:
            raise ValueError(method, " is a bad explanation method")

    @staticmethod
    def explain_method_as_int(method: str) -> int:
        if method.upper() == "SHAP":
            return ExplanationMethod.SHAP
        elif method.upper() == "LIME":
            return ExplanationMethod.LIME
        else:
            raise ValueError(method, " is a bad explanation method")



"""
Utils module for the antakia package.
"""
import math
import time
from enum import EnumMeta, Enum
from numbers import Number

import numpy as np
from functools import wraps

import pandas as pd
from pandas.api.types import is_bool_dtype, is_integer_dtype


def in_index(indexes: list, X: pd.DataFrame) -> bool:
    """
    Checks if a list of indexes is in the index of a DataFrame
    """
    try:
        X.loc[indexes]
        return True
    except KeyError:
        return False


def rows_to_mask(X: pd.DataFrame, rows_list: list) -> pd.Series:
    """
    Converts DataFrame row numbers to Index numbers
    """
    mask = pd.Series(np.zeros(len(X)), index=X.index)
    mask.iloc[rows_list] = 1
    return mask.astype(bool)


def indexes_to_rows(X: pd.DataFrame, indexes_list: list) -> list:
    """
    Converts DataFrame Index numbers to row numbers
    """
    index = pd.Series(np.arange(len(X)), index=X.index)
    return index.loc[indexes_list].tolist()


def mask_to_rows(mask: pd.Series) -> list:
    """
    converts a mask to row indices (i.e. iloc)
    """
    return mask_to_index(mask.reset_index(drop=True))


def mask_to_index(mask: pd.Series) -> list:
    """
    converts a mask to indices (i.e. loc)
    """
    return mask[mask].index.tolist()


def boolean_mask(X: pd.DataFrame, value: bool = True):
    """
    builds a constant series indexed on X with value as value
    """
    return pd.Series([value] * len(X), index=X.index)


def timeit(method):
    def timed(*args, **kw):
        ts = time.time()
        result = method(*args, **kw)
        te = time.time()
        if 'log_time' in kw:
            name = kw.get('log_name', method.__name__.upper())
            kw['log_time'][name] = int((te - ts) * 1000)
        else:
            print('%r  %2.2f ms' % (method.__name__, (te - ts) * 1000))
        return result

    return timed


def debug(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        print(f"Calling {func.__name__} with args: {args} and kwargs: {kwargs}")
        result = func(*args, **kwargs)
        print(f"{func.__name__} returned: {result}")
        return result

    return wrapper


def compute_step(min, max):
    """
    compute rounded min, max, and step values
    """
    step = (max - min) / 100
    round_value = round(math.log(step / 2) / math.log(10)) - 1
    step = np.round(step, -round_value)
    min_ = np.round(min, -round_value)
    max_ = int((max - min_) / step + 1) * step + min
    return min_, max_, step


def get_mask_comparison_color(rules_mask, selection_mask):
    """
    compute colors for comparison between two masks
    """
    colors_info = {
        'matched': 'blue',
        'error type 1': 'orange',
        'error type 2': 'red',
        'other data': 'grey'
    }
    color = pd.Series(index=selection_mask.index, dtype=str)
    color[selection_mask & rules_mask] = colors_info['matched']
    color[~selection_mask & rules_mask] = colors_info['error type 1']
    color[selection_mask & ~rules_mask] = colors_info['error type 2']
    color[~selection_mask & ~rules_mask] = colors_info['other data']
    return color, colors_info


# First color can't be blue, reserved for the rules - grey is reserved to background
colors = ["red", "blue", "green", "yellow", "orange", "pink", "brown", "cyan", "black"]


class MetaEnum(EnumMeta):
    def __contains__(cls, item):
        try:
            cls(item)
        except ValueError:
            return False
        return True


def format_data(x):
    if isinstance(x, str):
        return x[:15]
    if isinstance(x, Number):
        return format_number(x)
    if isinstance(x, pd.Series):
        if is_bool_dtype(x.dtype) or is_integer_dtype(x.dtype):
            return x
        return x.apply(format_data)
    return x


def format_number(value: Number) -> str:
    def format_str(value):
        if abs(value) < 1e-2 or abs(value) > 1e4:
            return '.2e'
        return '.2f'

    if value == 0:
        return value
    elif pd.isna(value):
        return 'NaN'
    return format(value, format_str(value))


class BaseEnum(Enum, metaclass=MetaEnum):
    pass


ProblemCategory = BaseEnum('ProblemCategory',
                           ['classification_with_proba', 'classification_proba', 'classification',
                            'classification_label_only', 'regression', 'auto'])



import pandas as pd
from sklearn.inspection import permutation_importance


class NotFittedError(Exception):
    pass


class MLModel:
    def __init__(self, model, name, fitted=False):
        self.fitted = fitted
        self.model = model
        self.name = name
        self.feature_importances_ = None

    def fit(self, X, y):
        if not self.fitted:
            res = self.model.fit(X, y)
            self.fitted = True
            return res

    def predict(self, X, *args, **kwargs):
        if self.fitted:
            pred = self.model.predict(X, *args, **kwargs)
            if isinstance(pred, (pd.DataFrame, pd.Series)):
                return pred
            else:
                return pd.Series(pred, index=X.index)
        raise NotFittedError()

    def compute_feature_importances(self, X, y, score, score_type):
        if self.feature_importances_ is None:
            if hasattr(self.model, 'feature_importances_'):
                self.feature_importances_ = pd.Series(
                    self.model.feature_importances_,
                    index=X.columns
                )
            else:
                def scorer(model, X, y):
                    y_pred = model.predict(X)
                    if score_type == 'minimize':
                        return -score(y, y_pred)
                    else:
                        return score(y, y_pred)

                fi = permutation_importance(
                    self.model, X, y, n_repeats=10, random_state=42, n_jobs=-1, scoring=scorer
                )
                self.feature_importances_ = pd.Series(
                    fi.importances_mean,
                    index=X.columns
                )

    def fit_and_compute_fi(self, X_train, y_train, X_test, y_test, score, score_type):
        self.fit(X_train, y_train)
        self.compute_feature_importances(X_test, y_test, score, score_type)


class AvgRegressionBaseline:
    def fit(self, X, y, *args, **kwargs):
        self.mean = y.mean()

    def predict(self, X, *args, **kwargs):
        return [self.mean] * len(X)


class AvgClassificationBaseline:
    def fit(self, X, y, *args, **kwargs):
        lst = list(y)
        self.majority_class = max(lst, key=lst.count)

    def predict(self, X, *args, **kwargs):
        return [self.majority_class] * len(X)


class LinearMLModel(MLModel):

    def fit(self, X, *args, **kwargs):
        super().fit(X, *args, **kwargs)
        self.means = X.mean()

    def global_explanation(self):
        coefs = pd.Series(self.model.coef_, index=self.model.features_names_in_)
        coefs['intercept'] = self.model.intercept_
        return {
            'type': 'table',
            'value': coefs
        }

    def local_explanation(self, x):
        coefs = pd.Series(self.model.coef_, index=self.model.features_names_in_)
        exp = coefs * x - coefs * self.means
        return {
            'type': 'table',
            'prior': self.predict(x),
            'value': exp
        }


class GAMMLMdel(MLModel):
    def global_explanation(self):
        coefs = pd.Series(self.model.coef_, index=self.model.features_names_in_)
        coefs['intercept'] = self.model.intercept_
        return {
            'type': 'table',
            'value': coefs
        }

    def local_explanation(self, x):
        coefs = pd.Series(self.model.coef_, index=self.model.features_names_in_)
        exp = coefs * x - coefs * self.means
        return {
            'type': 'table',
            'prior': self.predict(x),
            'value': exp
        }



import pandas as pd
from sklearn.base import TransformerMixin

from antakia_core.utils.long_task import LongTask


class DimReducMethod(LongTask):
    """
    Class that allows to reduce the dimensionality of the data.

    Attributes
    ----------
    dimreduc_method : int, can be PCA, TSNE etc.
    dimension : int
        Dimension reduction methods require a dimension parameter
        We store it in the abstract class
    """

    # Class attributes methods
    dim_reduc_methods = ['PCA', 'UMAP', 'PaCMAP']
    dimreduc_method = -1

    allowed_kwargs = []

    def __init__(
            self,
            dimreduc_method: int,
            dimreduc_model: type[TransformerMixin],
            dimension: int,
            X: pd.DataFrame,
            default_parameters: dict = None,
            progress_updated: callable = None,
    ):
        """
        Constructor for the DimReducMethod class.

        Parameters
        ----------
        dimreduc_method : int
            Dimension reduction methods among DimReducMethod.PCA, DimReducMethod.TSNE, DimReducMethod.UMAP or DimReducMethod.PaCMAP
            We store it here (not in implementation class)
        dimension : int
            Target dimension. Can be DIM_TWO or DIM_THREE
            We store it here (not in implementation class)
        X : pd.DataFrame
            Stored in LongTask instance
        progress_updated : callable
            Stored in LongTask instance
        """
        if not DimReducMethod.is_valid_dimreduc_method(dimreduc_method):
            if dimreduc_method == -1:
                print('warning - method is not yet supported')
            else:
                raise ValueError(
                    dimreduc_method, " is a bad dimensionality reduction method"
                )
        if not DimReducMethod.is_valid_dim_number(dimension):
            raise ValueError(dimension, " is a bad dimension number")

        self.dimreduc_method = dimreduc_method
        self.default_parameters = default_parameters
        self.dimension = dimension
        self.dimreduc_model = dimreduc_model
        # IMPORTANT : we set the topic as for ex 'PCA/2' or 't-SNE/3' -> subscribers have to follow this scheme
        LongTask.__init__(self, X, progress_updated)

    @classmethod
    def dimreduc_method_as_str(cls, method: int) -> str:
        if method is None:
            return None
        elif 0 < method <= len(cls.dim_reduc_methods):
            return cls.dim_reduc_methods[method - 1]
        else:
            raise ValueError(f"{method} is an invalid dimensionality reduction method")

    @classmethod
    def dimreduc_method_as_int(cls, method: str) -> int:
        if method is None:
            return None
        try:
            i = cls.dim_reduc_methods.index(method) + 1
            return i
        except ValueError:
            raise ValueError(f"{method} is an invalid dimensionality reduction method")

    @classmethod
    def dimreduc_methods_as_list(cls) -> list[int]:
        return list(map(lambda x: x + 1, range(len(cls.dim_reduc_methods))))

    @classmethod
    def dimreduc_methods_as_str_list(cls) -> list[str]:
        return cls.dim_reduc_methods.copy()

    @staticmethod
    def dimension_as_str(dim) -> str:
        if dim == 2:
            return "2D"
        elif dim == 3:
            return "3D"
        else:
            raise ValueError(f"{dim}, is a bad dimension")

    @classmethod
    def is_valid_dimreduc_method(cls, method: int) -> bool:
        """
        Returns True if it is a valid dimensionality reduction method.
        """
        return 0 <= method - 1 < len(cls.dim_reduc_methods)

    @staticmethod
    def is_valid_dim_number(dim: int) -> bool:
        """
        Returns True if dim is a valid dimension number.
        """
        return dim in [2, 3]

    def get_dimension(self) -> int:
        return self.dimension

    @classmethod
    def parameters(cls) -> dict[str, dict[str, any]]:
        return {}

    def compute(self, **kwargs) -> pd.DataFrame:
        self.publish_progress(0)
        kwargs['n_components'] = self.get_dimension()
        param = self.default_parameters.copy()
        param.update(kwargs)

        dim_red_model = self.dimreduc_model(**param)
        if hasattr(dim_red_model, 'fit_transform'):
            X_red = dim_red_model.fit_transform(self.X)
        else:
            dim_red_model.fit(self.X)
            X_red = dim_red_model.transform(self.X)
        X_red = pd.DataFrame(X_red)

        self.publish_progress(100)
        return X_red

    @classmethod
    def scale_value_space(cls, X: pd.DataFrame, y: pd.Series) -> pd.DataFrame:
        """
        Scale the values in X so that it's reduced and centered and weighted with mi
        """
        std = X.std()
        std[std == 0] = 1
        from sklearn.feature_selection import mutual_info_regression
        mi = mutual_info_regression(X, y)
        return (X - X.mean()) / std * mi



import time
from abc import ABC, abstractmethod

import pandas as pd


def dummy_progress(*args, **kwargs):
    pass

class LongTask(ABC):
    """
    Abstract class to compute long tasks, often in a separate thread.

    Attributes
    ----------
    X : dataframe
    progress_updated : an optional callback function to call when progress is updated
    start_time : float
    progress:int
    """

    def __init__(self, X: pd.DataFrame=None, progress_updated: callable = None):
        if X is None:
            raise ValueError("You must provide a dataframe for a LongTask")
        self.X = X
        if progress_updated is None:
            progress_updated = dummy_progress
        self.progress_updated = progress_updated
        self.start_time = time.time()

    def publish_progress(self, progress: int):
        if self.progress_updated:
            self.progress_updated(progress, time.time() - self.start_time)

    @abstractmethod
    def compute(self, **kwargs) -> pd.DataFrame:
        """
        Method to compute the long task and update listener with the progress.
        """
        pass



from typing import List

import pandas as pd
from sklearn.metrics import mean_squared_error, mean_absolute_error, r2_score, accuracy_score, f1_score, \
    precision_score, recall_score
from sklearn.model_selection import train_test_split

from joblib import Parallel, delayed

from antakia_core.compute.model_subtitution.classification_models import *
from antakia_core.compute.model_subtitution.regression_models import *
import re

from antakia_core.utils.utils import ProblemCategory


def pretty_model_name(model_name):
    return model_name.replace('_', ' ').title()


def reduce_name(model_name):
    parts = re.split('\W+', model_name)
    name = ''
    for part in parts:
        name += part[0].upper()
        for char in part[1:]:
            if char.isupper():
                name += char
    if len(name) == 1:
        return model_name[:2].capitalize()
    return name


class InterpretableModels:
    available_scores = {
        'MSE': (mean_squared_error, 'minimize'),
        'MAE': (mean_absolute_error, 'minimize'),
        'R2': (r2_score, 'maximize'),
        'ACC': (accuracy_score, 'maximize'),
        'ACCURACY': (accuracy_score, 'maximize'),
        'F1': (f1_score, 'maximize'),
        'precision'.upper(): (precision_score, 'maximize'),
        'recall'.upper(): (recall_score, 'maximize'),
    }
    customer_model_name = pretty_model_name('original_model')

    def __init__(self, custom_score):
        if callable(custom_score):
            self.custom_score_str = custom_score.__name__.upper()
            self.custom_score = custom_score
            self.score_type = 'compute'
        else:
            self.custom_score_str = custom_score.upper()
            self.custom_score = self.available_scores[custom_score.upper()][0]
            self.score_type = self.available_scores[custom_score.upper()][1]

        self.models = {}
        self.scores = {}
        self.perfs = pd.DataFrame()
        self.selected_model = None

    def _get_available_models(self, task_type) -> List[type[MLModel]]:
        if task_type == ProblemCategory.regression:
            return [LinearRegression, LassoRegression, RidgeRegression, GaM,
                    EBM, DecisionTreeRegressor, AvgBaselineModel]
        return [AvgClassificationBaselineModel, DecisionTreeClassifier, LogisticRegression]

    def _init_models(self, task_type):
        for model_class in self._get_available_models(task_type):
            model = model_class()
            if model.name not in self.models:
                self.models[pretty_model_name(model.name)] = model

    def _init_scores(self, customer_model, task_type):
        if self.score_type == 'compute':
            self._compute_score_type(customer_model, X_test, y_test)
        if task_type == ProblemCategory.regression:
            scores_list = ['MSE', 'MAE', 'R2']
        else:
            scores_list = ['ACC', 'F1', 'precision'.upper(), 'recall'.upper(), 'R2']
        self.scores = {
            score: self.available_scores[score] for score in scores_list
        }
        self.scores[self.custom_score_str] = (self.custom_score, self.score_type)

    def _train_models(self, X_train, y_train, X_test, y_test):
        Parallel(n_jobs=1)(
            delayed(model.fit_and_compute_fi)(X_train, y_train, X_test, y_test, self.custom_score, self.score_type) for
            model_name, model in
            self.models.items() if not model.fitted)

    def _compute_score_type(self, customer_model, X: pd.DataFrame, y: pd.Series):
        y_pred = customer_model.predict(X)
        s1 = self.custom_score(y_pred, y)
        s2 = self.custom_score(y.sample(len(y)).values, y.values)
        self.score_type = 'maximize' if s1 > s2 else 'minimize'

    def get_models_performance(
            self,
            customer_model,
            X_train: pd.DataFrame,
            y_train: pd.Series,
            X_test: pd.DataFrame | None,
            y_test: pd.Series | None,
            task_type='regression') -> pd.DataFrame:
        if isinstance(task_type, str):
            task_type = ProblemCategory[task_type]
        if len(X_train) <= 50 or len(X_train.T) >= len(X_train):
            return pd.DataFrame()
        if X_test is None or len(X_test) == 0:
            print('no test set provided, splitting train set')
            X_train, X_test, y_train, y_test = train_test_split(X_train, y_train, test_size=0.2)
        self._init_scores(customer_model, task_type)
        self._init_models(task_type)
        if customer_model is not None:
            self.models[self.customer_model_name] = MLModel(customer_model, self.customer_model_name, True)
            self.models[self.customer_model_name].compute_feature_importances(X_test, y_test, self.custom_score,
                                                                              self.score_type)
        self._train_models(X_train, y_train, X_test, y_test)
        for model_name, model in self.models.items():
            y_pred = model.predict(X_test)
            for score_name, score in self.scores.items():
                self.perfs.loc[model_name, score_name] = score[0](y_test, y_pred)

        self.perfs['delta'] = (self.perfs[self.custom_score_str] - self.perfs.loc[
            self.customer_model_name, self.custom_score_str]) * (2 * (self.score_type == 'minimize') - 1)

        def get_delta_color(delta):
            return 'red' if delta > 0.01 else 'green' if delta < -0.01 else 'orange'

        self.perfs['delta_color'] = self.perfs['delta'].apply(get_delta_color)
        return self.perfs.sort_values('delta', ascending=True)

    def select_model(self, model_name):
        self.selected_model = model_name

    def selected_model_str(self) -> str:
        perf = self.perfs.loc[self.selected_model]
        reduced_name = reduce_name(self.selected_model)
        display_str = f'{reduced_name} - {self.custom_score_str}:{perf[self.custom_score_str]:.2f} ({perf["delta"]:.2f})'
        return display_str

    def reset(self):
        self.models = {}
        self.scores = {}
        self.perfs = pd.DataFrame()
        self.selected_model = None


if __name__ == '__main__':
    df = pd.read_csv('../../../../antakia/data/california_housing.csv').set_index('Unnamed: 0')
    df = df.sample(len(df))
    limit = int(2000 / 0.8)
    df = df.iloc[:limit]
    split_row = int(len(df) * 0.8)
    df_train = df[:split_row]
    df_test = df[split_row:]
    X_train = df_train.iloc[:, :8]  # the dataset
    y_train = df_train.iloc[:, 9]  # the target variable
    X_test = df_test.iloc[:, :8]  # the dataset
    y_test = df_test.iloc[:, 9]  # the target variable
    InterpretableModels(mean_squared_error).get_models_performance(
        None, X_train, y_train, X_test, y_test,
        task_type='regression')



from __future__ import annotations
import pandas as pd

from antakia_core.data_handler.rule import Rule
from antakia_core.utils.utils import boolean_mask, mask_to_index
from antakia_core.utils.variable import Variable, DataVariables


class RuleSet:
    """
    set of rules
    """

    def __init__(self, rules: list[Rule] | RuleSet | None = None):

        if isinstance(rules, RuleSet):
            rules = rules.rules.values()
        self.rules = {}
        if rules:
            for rule in rules:
                self.rules[rule.variable] = rule

    def add(self, value: Rule):
        """
        add a new rule
        combines it with the existing rule is variable already used
        Parameters
        ----------
        value

        Returns
        -------

        """
        if value.variable in self.rules:
            self.rules[value.variable] &= value
        self.rules[value.variable] = value

    def replace(self, value: Rule):
        """
        add or replace (if variable is already used) a new rule
        Parameters
        ----------
        value

        Returns
        -------

        """
        self.rules[value.variable] = value

    def __len__(self):
        return len(self.rules)

    def __repr__(self):
        if not self.rules:
            return ""
        return " and ".join([rule.__repr__() for rule in self.rules.values()])

    def to_dict(self):
        if not self.rules:
            return []
        return [rule.to_dict() for rule in self.rules.values()]

    def copy(self):
        return RuleSet(self.rules.values())

    def get_matching_mask(self, X: pd.DataFrame) -> pd.Series:
        """
        get the mask of samples validating the rule
        Parameters
        ----------
        X: dataset to get the mask from

        Returns
        -------

        """
        res = boolean_mask(X, True)
        if self.rules is not None:
            for rule in self.rules.values():
                res &= rule.get_matching_mask(X)
        return res

    def get_all_masks(self, X: pd.DataFrame) -> list[pd.Series]:
        """
        returns the list of rules masks on X
        Parameters
        ----------
        X

        Returns
        -------

        """
        masks = []
        if self.rules is not None:
            for rule in self.rules.values():
                masks.append(rule.get_matching_mask(X))
        return masks

    def get_matching_indexes(self, X):
        """
        get the list indexes of X validating the rule
        Parameters
        ----------
        X

        Returns
        -------

        """
        res = self.get_matching_mask(X)
        return mask_to_index(res)

    @classmethod
    def sk_rules_to_rule_set(cls, skrules, variables: DataVariables):
        """
        transform skope rules to a RuleSet
        Parameters
        ----------
        skrules
        variables

        Returns
        -------

        """
        rules_info = skrules[0]
        precision, recall, __ = rules_info[1]
        f1 = precision * recall * 2 / (precision + recall)
        score_dict = {
            "precision": round(precision, 3),
            "recall": round(recall, 3),
            "f1": round(f1, 3),
        }
        rule_strings = rules_info[0].split(" and ")

        rule_list = RuleSet()
        for rule in rule_strings:
            rule_parts = rule.split(' ')

            variable = variables.get_var(rule_parts[0])
            if len(rule_parts) != 3:
                raise ValueError('Rule not recognized')
            includes = '=' in rule_parts[1]
            value = float(rule_parts[2])
            if '<' in rule_parts[1]:
                temp_rule = Rule(variable, max=value, includes_max=includes)
            else:
                temp_rule = Rule(variable, min=value, includes_min=includes)
            rule_list.add(temp_rule)

        return rule_list, score_dict

    def get_rule(self, var: Variable) -> Rule | None:
        """
        find a rule on the variable
        Parameters
        ----------
        var

        Returns
        -------

        """
        return self.rules.get(var)



from typing import List

import numpy as np
import pandas as pd
from pandas.api.types import is_numeric_dtype


class Variable:
    """
    Describes each X column or Y value

    col_index : int
        The index of the column in the dataframe i come from (ds or xds)
        #TODO : I shoudl code an Abstract class for Dataset and ExplanationDataset
    column_name : str
        How it should be displayed in the GUI
    descr : str
        A description of the variable
    type : str
        The type of the variable
    sensible : bool
        Wether the variable is sensible or not
    contiuous : bool
    lat : bool
    lon : bool
    """

    def __init__(
            self,
            col_index: int,
            column_name: str,
            type: str,
            unit: str = None,
            descr: str = None,
            critical: bool = False,
            continuous: bool = True,
            lat: bool = False,
            lon: bool = False,
            **kwargs  # to ignore unknown args in building object
    ):
        self.col_index = col_index
        self.column_name = column_name
        self.type = type
        self.unit = unit
        self.descr = descr
        self.critical = critical
        self.continuous = continuous
        self.lat = lat
        self.lon = lon

    @property
    def display_name(self):
        return self.column_name.replace('_', ' ')

    @staticmethod
    def guess_variables(X: pd.DataFrame) -> 'DataVariables':
        """
        guess Variables from the Dataframe values
        Returns a DataVariable object, with one Variable for each column in X.
        """
        variables = []
        for i, col in enumerate(X.columns):
            var = Variable(i, col, X.dtypes[col])
            if isinstance(col, str):
                if col.lower() in ["latitude", "lat"]:
                    var.lat = True
                if col.lower() in ["longitude", "long", "lon"]:
                    var.lon = True
            var.continuous = Variable.is_continuous(X[col])
            variables.append(var)
        return DataVariables(variables)

    @staticmethod
    def import_variable_df(df: pd.DataFrame) -> 'DataVariables':
        """
        Builds The DataVariable object from a descriptive DataFrame
        Returns a DataVariable object, with one Variable for each column in X.
        """

        if "col_index" not in df.columns:
            df['col_index'] = np.arange(len(df))
        if 'column_name' not in df.columns:
            df['column_name'] = df.index
            if is_numeric_dtype(df['column_name']):
                raise KeyError('column_name (index) column is mandatory and should be string')
        if 'type' not in df.columns:
            raise KeyError('type column is mandatory')
        variables = df.apply(lambda row: Variable(**row), axis=1).to_list()
        return DataVariables(variables)

    @staticmethod
    def import_variable_list(var_list: list) -> 'DataVariables':
        """
        Builds The DataVariable object from alist of dict
        Returns a DataVariable object, with one Variable for each column in X.

        """
        variables = []
        for i in range(len(var_list)):
            if isinstance(var_list[i], dict):
                item = var_list[i]
                if "col_index" in item and "column_name" in item and "type" in item:
                    var = Variable(**item)
                    variables.append(var)
                else:
                    raise ValueError(
                        "Variable must a list of {key:value} with mandatory keys : [col_index, column_name, type] and optional keys : [unit, descr, critical, continuous, lat, lon]"
                    )
        return DataVariables(variables)

    @staticmethod
    def is_continuous(serie: pd.Series) -> bool:
        """
        guess if a variable is continous from its values
        """
        # TODO : precise continuous definition
        id_first_true = (serie > 0).idxmax()
        id_last_true = (serie > 0)[::-1].idxmax()
        return all((serie > 0).loc[id_first_true:id_last_true] == True)

    def __repr__(self):
        """
        Displays the variable as a string
        """
        text = f"{self.display_name}, col#:{self.col_index}, type:{self.type}"
        if self.descr is not None:
            text += f", descr:{self.descr}"
        if self.unit is not None:
            text += f", unit:{self.unit}"
        if self.critical:
            text += ", critical"
        if not self.continuous:
            text += ", categorical"
        if self.lat:
            text += ", is lat"
        if self.lon:
            text += ", is lon"
        return text

    def __eq__(self, other):
        return (
                self.col_index == other.col_index and
                self.column_name == other.column_name and
                self.type == other.type and
                self.unit == other.unit and
                self.descr == other.descr and
                self.critical == other.critical and
                self.continuous == other.continuous and
                self.lat == other.lat and
                self.lon == other.lon
        )

    def __hash__(self):
        return hash(self.column_name)


class DataVariables:
    """
    collection of Variables
    """

    def __init__(self, variables: List[Variable]):
        self.variables = {var.column_name: var for var in variables}

    def __str__(self):
        text = ""
        for var in self.variables.values():
            text += str(var.col_index) + ") " + str(var) + "\n"
        return text

    def columns_list(self):
        """
        get column_name list
        """
        return list(self.variables.keys())

    def get_var(self, column_name: str):
        """
        get variable by column_name
        """
        return self.variables.get(column_name)

    def __len__(self):
        return len(self.variables)

    def __eq__(self, other):
        for i in self.variables.values():
            if i not in other.variables.values():
                return False
        for j in other.variables.values():
            if j not in self.variables.values():
                return False
        return True



from __future__ import annotations

import pandas as pd
from sklearn.neighbors import KNeighborsClassifier

from antakia_core.compute.model_subtitution.model_interface import InterpretableModels
from antakia_core.data_handler.rules import RuleSet
from antakia_core.utils.utils import colors, boolean_mask, format_number


class Region:
    """
    class to handle regions
    a region is defined either by a selection of point or by a set of rules
    """
    region_colors = colors

    def __init__(self, X, rules: RuleSet | None = None, mask: pd.Series | None = None, color=None, num=-1):
        """

        Parameters
        ----------
        X : base dataframe to use for the rule
        rules : list of rules
        mask : selected points
        color: region color, if not provided, auto assigned
        """
        self.X = X
        self.num = num
        self.rules = RuleSet(rules)
        if mask is None:
            # if no mask, compute it
            if rules is not None:
                self.mask = self.rules.get_matching_mask(X)
            else:
                self.mask = pd.Series([False] * len(X), index=X.index)
        else:
            self.mask = mask
        self._color = color
        self.validated = False
        self.auto_cluster = False

    @property
    def color(self):
        """
        get region color
        Returns
        -------

        """
        if self._color is None:
            return self.region_colors[(self.num - 1) % len(self.region_colors)]
        return self._color

    @color.setter
    def color(self, c):
        """
        set region color
        Parameters
        ----------
        c

        Returns
        -------

        """
        self._color = c

    @property
    def name(self):
        """
        get region name
        Returns
        -------

        """
        name = repr(self.rules)
        if self.auto_cluster:
            if name:
                name = 'AC: ' + name
            else:
                name = "auto-cluster"
        return name

    def to_dict(self) -> dict[str, str | int | None]:
        """
        get region as dict
        Returns
        -------

        """
        dict_form = {
            "Region": self.num,
            "Rules": self.name,
            "Average": None,
            "Points": self.mask.sum(),
            "% dataset": f"{round(self.mask.mean() * 100, 2)}%",
            "Sub-model": None,
            'color': self.color
        }
        return dict_form

    def num_points(self) -> int:
        """
        get the number of points on the region
        Returns
        -------

        """
        return self.mask.sum()

    def dataset_cov(self):
        """
        get Region's dataset coverage (% of points in the Region)
        Returns
        -------

        """
        return self.mask.mean()

    def validate(self):
        """
        set Region as validated
        will not be erased by auto clustering
        Returns
        -------

        """
        self.validated = True

    def update_rule_set(self, rule_set: RuleSet):
        self.rules = rule_set
        self.mask = self.rules.get_matching_mask(self.X)
        self.validated = False


class ModelRegion(Region):
    """
    supercharged Region with an explainable predictive model
    """

    def __init__(self, X, y, X_test, y_test, customer_model, rules: RuleSet | None = None,
                 mask: pd.Series | None = None, color=None,
                 score=None, num=-1):
        """

        Parameters
        ----------
        X: base train dataset
        y: relative target
        X_test: test dataset
        y_test: relative target
        customer_model: customer model
        rules: list of rules defining the region
        mask: mask defining the region
        color: region's color
        score: customer provided scoring method
        """
        super().__init__(X, rules, mask, color, num)
        self.y = y
        self.X_test = X_test
        self._test_mask = None
        self.y_test = y_test
        self.customer_model = customer_model
        self.interpretable_models = InterpretableModels(score)

    def to_dict(self):
        """
        transform region to dict
        Returns
        -------

        """
        dict_form = super().to_dict()
        if self.interpretable_models.selected_model is not None:
            dict_form['Sub-model'] = self.interpretable_models.selected_model_str()
        dict_form["Average"] = format_number(self.y[self.mask].mean())
        return dict_form

    def select_model(self, model_name: str):
        """
        select a model between all interpretable models
        Parameters
        ----------
        model_name : model to select

        Returns
        -------

        """
        self.interpretable_models.select_model(model_name)

    def train_substitution_models(self, task_type):
        """
        train substitution models
        Returns
        -------

        """
        if self.X_test is not None:
            self.interpretable_models.get_models_performance(
                self.customer_model,
                self.X.loc[self.mask],
                self.y.loc[self.mask],
                self.X_test.loc[self.test_mask],
                self.y_test.loc[self.test_mask],
                task_type=task_type
            )
        else:
            self.interpretable_models.get_models_performance(
                self.customer_model,
                self.X.loc[self.mask],
                self.y.loc[self.mask],
                None,
                None,
                task_type=task_type
            )

    @property
    def perfs(self):
        """
        get model performance statistics
        Returns
        -------

        """
        perfs = self.interpretable_models.perfs
        if len(perfs) == 0:
            return perfs
        return perfs.sort_values('delta', ascending=True)

    @property
    def delta(self):
        """
        get performance difference between selected model and customer model
        Returns
        -------

        """
        if self.interpretable_models.selected_model:
            return self.interpretable_models.perfs.loc[self.interpretable_models.selected_model, 'delta']
        return 0

    @property
    def test_mask(self):
        """
        select testing sample from test set
        Returns
        -------

        """
        if self._test_mask is None:
            if self.rules and len(self.rules) > 0:
                self._test_mask = self.rules.get_matching_mask(self.X_test)
            else:
                if not self.mask.any() or self.mask.all():
                    return boolean_mask(self.X_test, self.mask.mean())
                knn = KNeighborsClassifier().fit(self.X, self.mask)
                self._test_mask = pd.Series(knn.predict(self.X_test), index=self.X_test.index)
        return self._test_mask

    def get_model(self, model_name):
        return self.interpretable_models.models[model_name]

    def get_selected_model(self):
        if self.interpretable_models.selected_model is None:
            return None
        return self.get_model(self.interpretable_models.selected_model)

    def predict(self, X):
        mask = self.rules.get_matching_mask(X)
        model = self.get_selected_model()
        if model is not None:
            return model.predict(X[mask]).reindex(X.index)
        return pd.Series(index=X.index)

    def update_rule_set(self, rule_set: RuleSet):
        super().update_rule_set(rule_set)
        self.interpretable_models.reset()

class RegionSet:
    """
    group of regions
    """

    def __init__(self, X):
        """

        Parameters
        ----------
        X: reference dataset
        """
        self.regions: dict[int:Region] = {}
        self.insert_order: list[int] = []
        self.display_order: list[Region] = []
        self.X = X

    def get_new_num(self) -> int:
        """
        get a new Region id
        Returns
        -------

        """
        if len(self.regions) == 0:
            return 1
        else:
            for i in range(1, len(self.regions) + 1):
                if self.regions.get(i) is None:
                    return i
            return len(self.regions) + 1

    def get_max_num(self) -> int:
        """
        get biggest region id
        Returns
        -------

        """
        if not len(self.regions):
            return 0
        return max(self.insert_order)

    def add(self, region: Region) -> None:
        """
        add a new Region to the set
        prefer the add region method
        Parameters
        ----------
        region

        Returns
        -------

        """
        if region.num < 0:
            num = self.get_new_num()
            region.num = num
        if region.num in self.regions:
            self.remove(region.num)

        self.regions[region.num] = region
        self.insert_order.append(region.num)
        self.display_order.append(region)

    def add_region(self, rules: RuleSet = None, mask=None, color=None, auto_cluster=False) -> Region:
        """
        create a Region from a rule set or a mask
        Parameters
        ----------
        rules : rule list
        mask : selection mask
        color : region color
        auto_cluster: is from autoclustering ?

        Returns
        -------
        the created region

        """
        if mask is not None:
            mask = mask.reindex(self.X.index).fillna(False)
        region = Region(X=self.X, rules=rules, mask=mask, color=color)
        region.auto_cluster = auto_cluster
        self.add(region)
        return region

    def extend(self, region_set: RegionSet) -> None:
        """
        add the provided RegionSet into the current one
        rebuilds all Regions
        Parameters
        ----------
        region_set

        Returns
        -------

        """
        for region in region_set.regions.values():
            self.add_region(region.rules, region.mask, region._color, region.auto_cluster)

    def remove(self, region_num) -> None:
        """
        remove Region from set
        Parameters
        ----------
        region_num

        Returns
        -------

        """
        self.insert_order.remove(region_num)
        self.display_order.remove(self.regions[region_num])
        del self.regions[region_num]

    def to_dict(self) -> list[dict]:
        """
        dict like RegionSet
        Returns
        -------

        """
        return [region.to_dict() for region in self.display_order]

    def get_masks(self) -> list[pd.Series]:
        """
        returns all Region masks
        Returns
        -------

        """
        return [region.mask for region in self.display_order]

    @property
    def mask(self):
        """
        get the union mask of all regions
        Returns
        -------

        """
        union_mask = boolean_mask(self.X, False)
        for mask in self.get_masks():
            union_mask |= mask
        return union_mask

    def get_colors(self) -> list[str]:
        """
        get the list of Region colors
        Returns
        -------

        """
        return [region.color for region in self.display_order]

    def get_color_serie(self) -> pd.Series:
        """
        get a pd.Series with for each sample of self.X its region color
        the value is set to grey if the sample is not in any Region of the region set
        Returns
        -------

        """
        color = pd.Series(["grey"] * len(self.X), index=self.X.index)
        for region in self.display_order:
            color[region.mask] = region.color
        return color

    def __len__(self) -> int:
        """
        size of the region set
        Returns
        -------

        """
        return len(self.regions)

    def get(self, i) -> Region | None:
        """
        get a specific region by id
        Parameters
        ----------
        i

        Returns
        -------

        """
        return self.regions.get(i)

    def clear_unvalidated(self):
        """
        remove all unvalidated regions
        Returns
        -------

        """
        for i in list(self.regions.keys()):
            if not self.regions[i].validated:
                self.remove(i)

    def pop_last(self) -> Region:
        """
        removes and return the last region
        Returns
        -------

        """
        if len(self.insert_order) > 0:
            num = self.insert_order[-1]
            region = self.get(num)
            if not self.regions[num].validated:
                self.remove(num)
            return region

    def sort(self, by, ascending=True):
        """
        sort the region set by id, size, insert order
        Parameters
        ----------
        by : 'region_num'|'size'|'insert'
        ascending

        Returns
        -------

        """
        if by == 'region_num':
            key = lambda x: x.num
        elif by == 'size':
            key = lambda x: x.num_points()
        elif by == 'insert':
            key = lambda x: self.insert_order.index(x)
        self.display_order.sort(key=key, reverse=not ascending)

    def stats(self) -> dict:
        """ Computes the number of distinct points in the regions and the coverage in %
        """
        union_mask = self.mask
        stats = {
            'regions': len(self),
            'points': union_mask.sum(),
            'coverage': round(100 * union_mask.mean()),
        }
        return stats


class ModelRegionSet(RegionSet):
    """
    Supercharged RegionSet to handle interpretable models
    """

    def __init__(self, X, y, X_test, y_test, model, score):
        """

        Parameters
        ----------
        X: reference DatafFrame
        y: target series
        X_test: test set
        y_test: target test set
        model: customer model
        score: scoring method
        """
        super().__init__(X)
        self.y = y
        self.X_test = X_test
        self.y_test = y_test
        self.model = model
        self.score = score

    def add(self, region: Region):
        if not isinstance(region, ModelRegion):
            m_region = ModelRegion(
                X=self.X,
                y=self.y,
                X_test=self.X_test,
                y_test=self.y_test,
                customer_model=self.model,
                score=self.score,
                rules=region.rules,
                mask=region.mask,
                color=region._color,
                num=region.num
            )
            m_region.validated = region.validated
            region = m_region
        super().add(region)

    def add_region(self, rules: RuleSet = None, mask=None, color=None, auto_cluster=False) -> Region:
        """
        add new ModelRegion
        Parameters
        ----------
        rules
        mask
        color
        auto_cluster

        Returns
        -------

        """
        if mask is not None:
            mask = mask.reindex(self.X.index).fillna(False)
        region = ModelRegion(
            X=self.X,
            y=self.y,
            X_test=self.X_test,
            y_test=self.y_test,
            customer_model=self.model,
            score=self.score,
            rules=rules,
            mask=mask,
            color=color
        )
        region.num = -1
        region.auto_cluster = auto_cluster
        self.add(region)
        return region

    def get(self, i) -> ModelRegion | None:
        return super().get(i)

    def stats(self) -> dict:
        base_stats = super().stats()
        delta_score = 0
        for region in self.regions.values():
            weight = region.mask.sum()
            delta = region.delta
            delta_score += weight * delta
        delta_score /= len(self.X)
        base_stats['delta_score'] = delta_score
        return base_stats

    def predict(self, X):
        prediction = pd.Series(index=X.index)
        for region in self.regions.values():
            prediction = prediction.fillna(region.predict(X))
        return prediction



import pacmap
import pandas as pd
from sklearn.decomposition import PCA
from openTSNE import TSNE

from antakia_core.compute.dim_reduction.dim_reduc_method import DimReducMethod


# ===========================================================
#         Projections / Dim Reductions implementations
# ===========================================================


class PCADimReduc(DimReducMethod):
    """
    PCA computation class.
    """
    dimreduc_method = DimReducMethod.dimreduc_method_as_int('PCA')

    def __init__(self, X: pd.DataFrame, dimension: int = 2, callback: callable = None):
        super().__init__(
            self.dimreduc_method,
            PCA,
            dimension,
            X,
            progress_updated=callback,
            default_parameters={
                'n_components': dimension,
            }
        )
        self.allowed_kwargs = ['copy', 'whiten', 'svd_solver', 'tol', 'iterated_power', 'n_oversamples',
                               'power_iteration_normalizer', 'random_state']


class TSNEwrapper(TSNE):
    def fit_transform(self, X):
        return pd.DataFrame(self.fit(X.values), index=X.index)


class TSNEDimReduc(DimReducMethod):
    """
    T-SNE computation class.
    """
    dimreduc_method = -1  # DimReducMethod.dimreduc_method_as_int('TSNE')

    def __init__(self, X: pd.DataFrame, dimension: int = 2, callback: callable = None):
        super().__init__(
            self.dimreduc_method,
            TSNEwrapper,
            dimension,
            X,
            progress_updated=callback,
            default_parameters={
                'n_components': dimension,
                'n_jobs': -1
            }
        )
        self.allowed_kwargs = ['perplexity', 'early_exaggeration', 'learning_rate', 'n_iter',
                               'n_iter_without_progress', 'min_grad_norm', 'metric', 'metric_params', 'init', 'verbose',
                               'random_state', 'method', 'angle', 'n_jobs'
                               ]

    @classmethod
    def parameters(cls) -> dict:
        return {
            'perplexity': {
                'type': float,
                'min': 5,
                'max': 50,
                'default': 12
            },
            'learning_rate': {
                'type': [float, str],
                'min': 10,
                'max': 1000,
                'default': 'auto'
            }
        }


class UMAPDimReduc(DimReducMethod):
    """
    UMAP computation class.
    """
    dimreduc_method = DimReducMethod.dimreduc_method_as_int('UMAP')

    def __init__(self, X: pd.DataFrame, dimension: int = 2, callback: callable = None):
        import umap
        super().__init__(
            self.dimreduc_method,
            umap.UMAP,
            dimension,
            X,
            progress_updated=callback,
            default_parameters={
                'n_components': dimension,
                'n_jobs': -1
            }
        )
        self.allowed_kwargs = ['n_neighbors', 'metric', 'metric_kwds', 'output_metric',
                               'output_metric_kwds', 'n_epochs', 'learning_rate', 'init', 'min_dist', 'spread',
                               'low_memory', 'n_jobs', 'set_op_mix_ratio', 'local_connectivity', 'repulsion_strength',
                               'negative_sample_rate', 'transform_queue_size', 'a', 'b', 'random_state',
                               'angular_rp_forest', 'target_n_neighbors', 'target_metric', 'target_metric_kwds',
                               'target_weight', 'transform_seed', 'transform_mode', 'force_approximation_algorithm',
                               'verbose', 'tqdm_kwds', 'unique', 'densmap', 'dens_lambda', 'dens_frac',
                               'dens_var_shift', 'output_dens', 'disconnection_distance', 'precomputed_knn',
                               ]

    @classmethod
    def parameters(cls) -> dict:
        return {
            'n_neighbors': {
                'type': int,
                'min': 1,
                'max': 200,
                'default': 15
            },
            'min_dist': {
                'type': float,
                'min': 0.1,
                'max': 0.99,
                'default': 0.1
            }
        }


class PaCMAPDimReduc(DimReducMethod):
    """
    PaCMAP computation class.

    """
    dimreduc_method = DimReducMethod.dimreduc_method_as_int('PaCMAP')

    def __init__(self, X: pd.DataFrame, dimension: int = 2, callback: callable = None):
        super().__init__(
            self.dimreduc_method,
            pacmap.PaCMAP,
            dimension,
            X,
            progress_updated=callback,
            default_parameters={
                'n_components': dimension,
            }
        )
        self.allowed_kwargs = ['n_neighbors', 'MN_ratio', 'FP_ratio', 'pair_neighbors', 'pair_MN',
                               'pair_FP', 'distance', 'lr', 'num_iters', 'apply_pca', 'intermediate',
                               'intermediate_snapshots', 'random_state']

    @classmethod
    def parameters(cls) -> dict:
        return {
            'n_neighbors': {
                'type': int,
                'min': 1,
                'max': 200,
                'default': 15
            },
            'MN_ratio': {
                'type': float,
                'min': 0.1,
                'max': 10,
                'default': 0.5,
                'scale': 'log'
            },
            'FP_ratio': {
                'type': float,
                'min': 0.1,
                'max': 10,
                'default': 2,
                'scale': 'log'
            }
        }


dim_reduc_factory: dict[int, DimReducMethod] = {
    dm.dimreduc_method: dm for dm in [PCADimReduc, TSNEDimReduc, UMAPDimReduc, PaCMAPDimReduc]
}


def compute_projection(X: pd.DataFrame, y: pd.Series, dimreduc_method: int, dimension: int,
                       progress_callback: callable = None,
                       **kwargs) -> pd.DataFrame:
    if not DimReducMethod.is_valid_dimreduc_method(dimreduc_method) or not DimReducMethod.is_valid_dim_number(
            dimension):
        raise ValueError("Cannot compute proj method #", dimreduc_method, " in ", dimension, " dimensions")

    X_scaled = DimReducMethod.scale_value_space(X, y)

    dim_reduc = dim_reduc_factory.get(dimreduc_method)
    default_kwargs = {'random_state': 9}
    default_kwargs.update(kwargs)
    dim_reduc_kwargs = {k: v for k, v in default_kwargs.items() if k in dim_reduc.allowed_kwargs}
    proj_values = pd.DataFrame(dim_reduc(X_scaled, dimension, progress_callback).compute(**dim_reduc_kwargs).values,
                               index=X.index)
    return proj_values



import warnings

import pandas as pd
from skope_rules_temp import SkopeRules

from antakia_core.utils.variable import Variable, DataVariables
from antakia_core.data_handler.rules import Rule, RuleSet


def skope_rules(
        df_mask: pd.Series,
        base_space_df: pd.DataFrame,
        variables: DataVariables = None,
        precision: float = 0.7,
        recall: float = 0.7,
        random_state=42
) -> (RuleSet, dict[str, float]):
    """
    variables : list of Variables of the app
    df_indexes : list of (DataFrame) indexes for the points selected in the GUI
    base_space_df : the dataframe on which the rules will be computed / extracted. May be VS or ES values
    precision for SKR binary classifer : defaults to 0.7
    recall for SKR binary classifer : defaults to 0.7
    """
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        # We convert df_indexes in row_indexes
        y_train = df_mask.astype(int)
        if variables is None:
            variables = Variable.guess_variables(base_space_df)

        sk_classifier = SkopeRules(
            feature_names=variables.columns_list(),
            random_state=random_state,
            n_estimators=5,
            recall_min=recall,
            precision_min=precision,
            max_depth_duplication=0,
            max_samples=1.0,
            max_depth=3,
        )

        sk_classifier.fit(base_space_df, y_train)

    if sk_classifier.rules_ != []:
        rules_list, score_dict = RuleSet.sk_rules_to_rule_set(sk_classifier.rules_, variables)
        return rules_list, score_dict

    else:
        return RuleSet(), {}
